package CASObservatory;

import ij.*;
import ij.process.*;
import ij.gui.*;
import java.awt.*;
import java.awt.geom.*;
import ij.plugin.*;
import ij.measure.ResultsTable;

public class Concentric_Circles implements PlugIn, DialogListener {
	private static int n = 5;
	private static double lineWidth = 2;
	private static double xcenter, ycenter;
	private static double innerR, outerR;
	private boolean measure;
	private boolean hide;
	private ImagePlus img;

	public void run(String arg) {
		if (IJ.versionLessThan("1.47i"))
			return;
		img = IJ.getImage();
		
	
		
		Roi roi = img.getRoi();
		Rectangle b = new Rectangle(0, 0, img.getWidth(), img.getHeight());
		boolean isRoi = roi!=null && roi.isArea();
		boolean isLine = roi!=null && roi.getType()==Roi.LINE;
		if (isRoi||isLine) b = roi.getBounds();
		if (outerR==0f || isRoi || isLine)
			setup(b, isLine);
		if (!IJ.isMacro())
			drawCircles();
		showDialog();
	}

	void drawCircles() {
		if (hide || n==0)
			img.setOverlay(null);
		else {
			Overlay overlay = new Overlay();
			double inc = (outerR-innerR)/(n-1);
			for (double r=innerR; r<=outerR+inc/100.0; r+=inc) {
				Roi circle = new OvalRoi(xcenter-r, ycenter-r, r*2, r*2);
				circle.setStrokeColor(Color.red);
				if (lineWidth>1)
					circle.setStrokeWidth(lineWidth);
				overlay.add(circle);
				if (n==1) break;
			}
		
			img.setOverlay(overlay);
			if (measure)
				measure(overlay);
		}
	}

	void setup(Rectangle b, boolean isLine) {
		xcenter = b.x + b.width/2;
		ycenter = b.y + b.height/2;
		float size = Math.min(b.width, b.height);
		if (isLine)
			size = (float)Math.sqrt(b.width*b.width+b.height*b.height);
		innerR = size/20f;
		outerR = size/2f;
	}

	void showDialog() {
		NonBlockingGenericDialog gd = new NonBlockingGenericDialog("Circles");
	//	GenericDialog gd = new GenericDialog("Circles");
		gd.addNumericField("Circles:", n, 0);
		gd.addNumericField("Line width:", lineWidth, 1);
		gd.addNumericField("X center:", xcenter, 1);
		gd.addNumericField("Y center:", ycenter, 1);
		gd.addNumericField("Inner radius:", innerR, 1);
		gd.addNumericField("Outer radius:", outerR, 1);
		gd.addCheckbox("Measure", false);
		gd.addCheckbox("Hide", false);
		gd.addDialogListener(this);
		gd.showDialog();
		if (gd.wasCanceled())
			img.setOverlay(null);
	}

	public boolean dialogItemChanged(GenericDialog gd, AWTEvent e) {
		if (gd.wasCanceled())
			return false;
		n = (int)gd.getNextNumber();
		lineWidth = (float)gd.getNextNumber();
		xcenter = (float)gd.getNextNumber();
		ycenter = (float)gd.getNextNumber();
		innerR = (float)gd.getNextNumber();
		outerR = (float)gd.getNextNumber();
		measure = gd.getNextBoolean();
		hide = gd.getNextBoolean();
		drawCircles();
		return true;
	}

	void measure(Overlay overlay) {
		int n = overlay.size();
		ResultsTable rt =  ResultsTable.getResultsTable();
		ImageProcessor ip = img.getProcessor();
		for (int i=0; i<n; i++) {	
			Roi circle = overlay.get(i);
			Rectangle2D.Double r = circle.getFloatBounds();
			FloatPolygon p = circle.getInterpolatedPolygon(1, false);
			Double sum = 0.0;
			for (int j=0; j<p.npoints; j++)
				sum += ip.getPixelValue((int)Math.round(p.xpoints[j]),(int)Math.round(p.ypoints[j]));
			double mean = sum/p.npoints;
			rt.incrementCounter();
			rt.addValue("Radius", r.width/2);
			rt.addValue("Samples", p.npoints);
			rt.addValue("Mean", mean);
		}
		rt.show("Results");
	}

}

